//OpenCV
#include <cv.h>

#include "MatrixTransformations.h"

using namespace cv;

void MatrixTransformations:: ResetTransform(double *transf){
    for (int i=0;i<16;i++) transf[i]=0.0;
    transf[0]=transf[5]=transf[10]=transf[15]=1.0;
}

void MatrixTransformations:: FromQuaternionToRotationMatrix(double qx,double qy,double qz, double qw, double *rot){
  rot[0]=pow(qw,2.0) + pow(qx,2.0) - pow(qy,2.0) - pow(qz,2.0);
  rot[1]=2.0*qx*qy - 2.0*qw*qz;
  rot[2]=2.0*qx*qz + 2.0*qw*qy;
  rot[3]=2.0*qx*qy + 2.0*qw*qz;
  rot[4]=pow(qw,2.0) - pow(qx,2.0) + pow(qy,2.0) - pow(qz,2.0);
  rot[5]=2.0*qy*qz - 2.0*qw*qx;
  rot[6]=2.0*qx*qz - 2.0*qw*qy;
  rot[7]=2.0*qy*qz + 2.0*qw*qx;
  rot[8]=pow(qw,2.0) - pow(qx,2.0) - pow(qy,2.0) + pow(qz,2.0);
}

void MatrixTransformations:: setRotMatrixInTransf(double *rot,double *transf){
    transf[0]=rot[0];
    transf[1]=rot[1];
    transf[2]=rot[2];
    transf[4]=rot[3];
    transf[5]=rot[4];
    transf[6]=rot[5];
    transf[8]=rot[6];
    transf[9]=rot[7];
    transf[10]=rot[8];
}

//Update the total odometry pose with a new increment of pose
void MatrixTransformations:: updateTransform(double *transf,double *incr_transf){
    double transf_res[16];

    CvMat incr_TRANSF = cvMat(4,4,CV_64FC1, &incr_transf[0]); //64F means double
    CvMat transf_NOW = cvMat(4,4,CV_64FC1, &transf[0]);
    CvMat transf_RES = cvMat(4,4,CV_64FC1, &transf_res[0]);

    cvGEMM(&transf_NOW,&incr_TRANSF, 1.0, NULL, 0.0, &transf_RES, 0);
    for (int i=0;i<16;i++) transf[i]=transf_res[i];
}

void MatrixTransformations:: getRotMatrixFromTransf(double *transf,double *rot){
    rot[0]=transf[0];
    rot[1]=transf[1];
    rot[2]=transf[2];
    rot[3]=transf[4];
    rot[4]=transf[5];
    rot[5]=transf[6];
    rot[6]=transf[8];
    rot[7]=transf[9];
    rot[8]=transf[10];
}

void MatrixTransformations:: FromRotationToQuaternion(double *rot,double *qx,double *qy, double *qz,double *qw ){
  double trace = rot[0] + rot[4] + rot[8];
  if( trace > 0 ) {// I changed M_EPSILON to 0
    double s = 0.5f / sqrtf(trace+ 1.0f);
    *qw = 0.25f / s;
    *qx = ( rot[7] - rot[5] ) * s;
    *qy = ( rot[2] - rot[6] ) * s;
    *qz = ( rot[3] - rot[1] ) * s;
  } else {
    if ( (rot[0] > rot[4]) && (rot[0] > rot[8]) ) {
      double s = 2.0f * sqrtf( 1.0f + rot[0] - rot[4] - rot[8]);
      *qw = (rot[7] - rot[5] ) / s;
      *qx = 0.25f * s;
      *qy = (rot[1] + rot[3] ) / s;
      *qz = (rot[2] + rot[6] ) / s;
    } else if (rot[4] > rot[8]) {
      double s = 2.0f * sqrtf( 1.0f + rot[4] - rot[0] - rot[8]);
      *qw = (rot[2] - rot[6] ) / s;
      *qx = (rot[1] + rot[3] ) / s;
      *qy = 0.25f * s;
      *qz = (rot[5] + rot[7] ) / s;
    } else {
      double s = 2.0f * sqrtf( 1.0f + rot[8] - rot[0] - rot[4] );
      *qw = (rot[3] - rot[1] ) / s;
      *qx = (rot[2] + rot[6] ) / s;
      *qy = (rot[5] + rot[7] ) / s;
      *qz = 0.25f * s;
    }
  }
}


void MatrixTransformations:: FromRotationMatrixToEulerYXZ(double *rot,double *p,double *y,double *r){
  *p=-asin(rot[5]);
  if (fabs(*p)<M_PI/2.0){
    *r=atan2(rot[3],rot[4]);
    *y=atan2(rot[2],rot[8]);
  }
}

void MatrixTransformations:: FromEulerToQuaternion(double pch, double yw, double rll,double *qx,double *qy,double *qz, double *qw){
    double c1,c2,c3,s1,s2,s3;

    c1 = cos(yw/2.0);
    c2 = cos(rll/2.0);
    c3 = cos(pch/2.0);
    s1 = sin(yw/2.0);
    s2 = sin(rll/2.0);
    s3 = sin(pch/2.0);

    *qw = c1*c2*c3 - s1*s2*s3;
    *qx = s1*s2*c3 +c1*c2*s3;
    *qy = s1*c2*c3 + c1*s2*s3;
    *qz = c1*s2*c3 - s1*c2*s3;
}

void MatrixTransformations:: FromEulerToRotationMatrixYXZ(double p,double y,double r, double *rot){
    double sx,sy,sz,cx,cy,cz;

    cx=cos(p);
    sx=sin(p);
    cy=cos(y);
    sy=sin(y);
    cz=cos(r);
    sz=sin(r);

    rot[0]=cy*cz+sx*sy*sz;
    rot[1]=sx*sy*cz-cy*sz;
    rot[2]=cx*sy;
    rot[3]=cx*sz;
    rot[4]=cx*cz;
    rot[5]=-sx;
    rot[6]=sx*cy*sz-sy*cz;
    rot[7]=sx*cy*cz+sy*sz;
    rot[8]=cx*cy;
}
