/**
  @file robot.h
  @brief header definitions
  @author Salvador Dominguez
  @date 13/8/2013
  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>

//YAML reading
#include <fstream>
#include "yaml-cpp/yaml.h"
#include <vector>

//Matrix tansformations
#include "MatrixTransformations.h"

#include <ros/ros.h>
#include <sensor_msgs/Joy.h>
#include <std_msgs/String.h>
#include <std_msgs/Float64.h>
#include <std_msgs/Bool.h>
#include <geometry_msgs/Twist.h>
#include <geometry_msgs/PoseWithCovarianceStamped.h>

#include "gpmrc_node/gpmrc_state.h"
#include "gpmrc_node/gpmrc_command.h"
#include "gpmrc_node/AxisData.h"
#include "eusebio/RobotStatus.h"

//TF
#include <geometry_msgs/TransformStamped.h>
#include <tf/transform_broadcaster.h>
#include <tf/transform_listener.h>

#include <sensor_msgs/LaserScan.h>
#include <sensor_msgs/PointCloud2.h>




//Visualization
#include <visualization_msgs/Marker.h>

//WiiIR sensor position
#define WII_IR_LONG_POS 0.495
#define WII_IR_HEIGHT 1.835

//Laser range finder sensor position
#define LASER_LONG_POS 0.4
#define LASER_HEIGHT_POS 0.1

//Joystick GPMRC
#define JOY_BUTT_1_LEFT     10
#define JOY_BUTT_2_LEFT     11
#define JOY_BUTT_1_RIGHT    8
#define JOY_BUTT_2_RIGHT    9
#define JOY_BUTT_UP         0
#define JOY_BUTT_DOWN       1
#define JOY_BUTT_LEFT       2
#define JOY_BUTT_RIGHT      3
#define JOY_BUTT_SQUARE     7
#define JOY_BUTT_CIRCLE     6
#define JOY_BUTT_TRIANGLE   4
#define JOY_BUTT_CROSS      5
#define JOY_BUTT_SEL        12
#define JOY_BUTT_START      13

//Joystick PC
#define MIN_NUM_JOY_AXES 6
#define MIN_NUM_JOY_BUTTONS 13

#define JOY_PC_BUTT_SQUARE     	0
#define JOY_PC_BUTT_CROSS      	1
#define JOY_PC_BUTT_CIRCLE     	2
#define JOY_PC_BUTT_TRIANGLE   	3
#define JOY_PC_BUTT_R1         	5
#define JOY_PC_BUTT_L2         	6
#define JOY_PC_BUTT_R2         	7
#define JOY_PC_BUTT_SEL        	8
#define JOY_PC_BUTT_START      	9
#define JOY_PC_BUTT_HOME	12	

#define JOY_PC_AXIS_X1               0
#define JOY_PC_AXIS_Y1               1
#define JOY_PC_BUTT_LEFT_RIGHT       4
#define JOY_PC_BUTT_UP_DOWN          5

#define VEL_JOY_LIN_COEF 800.0
#define VEL_JOY_ANG_COEF 440.0
#define VEL_OFFSET 0.1

//Point maneouver type (Joystick)
#define PASS_POINT_TYPE 		0
#define ORIENTATION_POINT_TYPE 		1
#define CHARGER_CONNECT_POINT_TYPE 	2
#define CHARGER_DISCONNECT_POINT_TYPE 	3
#define THROUGHT_DOOR_POINT_TYPE 	4

//Robot status
#define ROBOT_CHARGING 			0
#define ROBOT_CONNECTED_NOT_CHARGING 	1
#define ROBOT_NOT_CONNECTED 		2
#define ROBOT_NAVIGATING 		3
#define ROBOT_OBSTACULIZED 		4

//Speed  estimation
#define K_SPEED_COEF 0.1 //To filter the increments of linear movement to estimate the speed of the robot 

/*
#define MAX_LASER_BUMPER_DISTANCE 0.8
#define MIN_LASER_BUMPER_DISTANCE 0.3
#define LASER_BUMPER_ANGLE_RANGE (M_PI/4.0)
*/
#define SAFE_DIST_RANGE_FRONT 0.50
#define SAFE_DIST_RANGE_LAT   0.30
//Period of forced charger pose setting.
#define SET_CHARGER_POSE_PERIOD 10.0

//GPMRC Digital Status
#define DS_CONTR_MAN_AUTO 0
#define DS_CHARGER_ACTIVE 1
#define DS_CONTR_CHARGER_CONNECTED 2

#define MANUAL_CONTROL true
#define AUTO_CONTROL false


extern eusebio::RobotStatus robot_st;

using namespace std;
//using namespace message_filters;


//Class for processing of the incoming images
class Robot
{
public:
  double correction_transf[16];
  double global_transf[16];
  float fps;

  //The robot status ROS topic
  eusebio::RobotStatus robot_st;
  double qx,qy,qz,qw;
  //double pitch,roll,yaw;
  ros::Time timestamp;

  ros::Publisher pub_gpmrc_command;
  ros::Publisher pub_last_scan;
  ros::Publisher pub_robot_status;
  ros::Publisher pub_ch_pose;
  ros::Publisher pub_obstacle;
  ros::Publisher pub_bumper_scan;
  ros::Publisher pub_marker;

  MatrixTransformations mt;

  bool new_joy_data;
  bool new_remote_data;
  bool dead_man_button;
  bool sel_button;
  bool start_button;
  bool circle_button;
  bool cross_button;
  bool square_button;
  bool triangle_button;
  bool home_button;
  bool R2_button;
  bool R1_button;
  bool bmp_front;
  bool bmp_back;
  int sonar_front;
  int sonar_back;
  double ref_joy_x,ref_joy_y;
  float ref_ext_lin,ref_ext_ang;
  bool charger_detected;
  double pls_dist;
  double incr_dist_robot,incr_t,incr_ang;

  double incr_x,incr_y;
  double laser_long_pos,laser_height_pos,laser_angle_pos;
  double dist_between_wheels;

  bool steering_calibrated;
  double wii_ir_long_pos;
  double wii_ir_height;
  int lin_act_state;
  double max_vel_pls;
  sensor_msgs::LaserScan last_scan;

  float linear_speed,ref_steer,ref_vel,steering_angle;

  //Frame identifiers
  string base_frame_name,odom_frame_name,map_frame_name,laser_frame_name,base_rear_axis_name,base_center_name,wii_ir_frame_name,mark_frame_name;

  double charging_pose_x, charging_pose_y, charging_pose_qx, charging_pose_qy, charging_pose_qz, charging_pose_qw;

  float robot_length,robot_width,robot_height;//Robot body dimensions
  double xs_odom,ys_odom,theta_odom;

  int incr_pls_left_last,incr_pls_right_last;
  Robot(){
      odom_frame_name="/odom";
      map_frame_name="/map";
      base_frame_name="/base_link";
      laser_frame_name="/laser";
      base_rear_axis_name="/base_rear_axis";
      base_center_name="/base_center";
      wii_ir_frame_name="/wii_ir_frame";
      new_joy_data=false;
      new_remote_data=false;
      dead_man_button=false;
      sel_button=false;
      start_button=false;
      circle_button=false;
      cross_button=false;
      square_button=false;
      triangle_button=false;
      home_button=false;
      bmp_front=false;
      bmp_back=false;
      sonar_front=0; //0=NO obstacle, 1=front, 2=left, 3=right
      sonar_back=0;
      ref_joy_x=0.0,ref_joy_y=0.0;
      ref_ext_lin=0.0,ref_ext_ang=0.0;
      pls_dist=100.0;
      dist_between_wheels=0.4;
      wii_ir_long_pos=WII_IR_LONG_POS;
      wii_ir_height=WII_IR_HEIGHT;
      max_vel_pls=30.0;
      laser_long_pos=LASER_LONG_POS;
      laser_height_pos=LASER_HEIGHT_POS;
      laser_angle_pos=0.0;
      charger_detected=false;
      mark_frame_name="/mark_frame";
      robot_length=robot_width=robot_height=0.5;
      robot_st.robotID="EUSEBIO-V1.0";
      robot_st.theta=0.0;
      robot_st.xs=0.0;
      robot_st.ys=0.0;
      mt.ResetTransform(&correction_transf[0]);
      mt.ResetTransform(&global_transf[0]);
      xs_odom=ys_odom=theta_odom=0.0;
      incr_pls_left_last=incr_pls_right_last=0;
      linear_speed=ref_steer=ref_vel=steering_angle=0.0;
  }

  //Callbacks for topics
  void CtrlRobot(int16_t reflin, int16_t refang, std::vector<unsigned char> buttons);
  void TwistToTracks(double ref_x,double ref_y,int16_t *ref1,int16_t *ref2);
  void ReadConfigurationFile(char *filename);
  void ReadRobotStatus(eusebio::RobotStatus *rs,string filename);
  void WriteRobotStatus(eusebio::RobotStatus rs, string filename);
  void ReadChargingPose(string filename);
  void SetChargingPose(void);
  void PublishOdom6d(void);
  void PublishChargerMarker(unsigned int status);
  void PublishRobotMarker(void);
};
