///////////////////////////////////////////////////////////////////////////////
// The serial_port package provides small, simple static libraries to access 
// serial devices
//
// Copyright (C) 2008, Morgan Quigley
//
// Redistribution and use in source and binary forms, with or without 
// modification, are permitted provided that the following conditions are met:
//   * Redistributions of source code must retain the above copyright notice, 
//     this list of conditions and the following disclaimer.
//   * Redistributions in binary form must reproduce the above copyright 
//     notice, this list of conditions and the following disclaimer in the 
//     documentation and/or other materials provided with the distribution.
//   * Neither the name of Stanford University nor the names of its 
//     contributors may be used to endorse or promote products derived from 
//     this software without specific prior written permission.
//   
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
// POSSIBILITY OF SUCH DAMAGE.

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>


#include "serial_port/lightweightserial.h"

#define READ_TIMEOUT          500000      /* less than 1e6 */
#define _POSIX

LightweightSerial::LightweightSerial(char *port, unsigned int baud,void (*handler)(int)):
  baud(baud), fd(0), happy(false)
{
  serial_connect(port, handler);
  serial_configure(baud,'N');
  return;
}

LightweightSerial::~LightweightSerial()
{
  if (happy==true)    close(fd);
  fd = 0; // prevent future reads...
}

int LightweightSerial::serial_connect(char *port, void (*handler)(int))
{
  struct termios newtio;
  struct sigaction saio;           /* definition of signal action */

  //fd = open(port, O_RDWR | O_SYNC | O_NOCTTY | O_NONBLOCK , 0);
  fd = open(port, O_RDWR | O_SYNC | O_NOCTTY , 0);

  if(fd < 0) {
    fprintf(stderr,"Serial I/O Error:  Could not open port %s\n", port);
    happy=false;
    return -1;
  }

  if (handler!=NULL){
    /* install the signal handler before making the device asynchronous */
    sigemptyset(&saio.sa_mask);
    saio.sa_handler = handler;
    saio.sa_flags = 0;
    saio.sa_restorer = NULL;
    sigaction(SIGIO,&saio,NULL);

    /* allow the process to receive SIGIO */
    fcntl(fd, F_SETOWN, getpid());
    /* Make the file descriptor asynchronous (the manual page says only O_APPEND and O_NONBLOCK, will work with F_SETFL...) */
    fcntl(fd, F_SETFL, FASYNC);
  }

  tcgetattr(fd, &newtio);

  bzero(&newtio, sizeof(newtio)); /* clear struct for new port settings */
  newtio.c_cflag = B115200 | CS8 | CLOCAL | CREAD;
  newtio.c_iflag = IGNPAR ;
  newtio.c_oflag = 0;

  newtio.c_cc[VTIME] = 0;
  newtio.c_cc[VMIN] = 1;

  tcflush(fd, TCIFLUSH);
  tcsetattr(fd, TCSANOW, &newtio);

  happy=true;
  printf("We are happy.The port is opened sucessfully\n");
  return 0;
}

void LightweightSerial::serial_configure(unsigned int baudrate, char parity)
{
  switch(baudrate) {
  case 9600:
    serial_setparms("9600", parity, "8", 0, 0);
    break;
  case 19200:
    serial_setparms("19200", parity, "8", 0, 0);
    break;
  case 38400:
    serial_setparms("38400", parity, "8", 0, 0);
    break;
  case 57600:
  case 55555:
    serial_setparms("57600", parity, "8", 0, 0);
    break;
  case 115200:
    serial_setparms("115200", parity, "8", 0, 0);
  case 500000:
    serial_setparms("500000", parity, "8", 0, 0);
    break;
  }
}

void LightweightSerial::serial_disconnect(void)
{
  if (happy==true) {
    close(fd);
  }
  return;
}


void LightweightSerial::serial_setrts(void)
{
#if defined(TIOCM_RTS) && defined(TIOCMODG)
  int mcs = 0;

  ioctl(fd, TIOCMODG, &mcs);
  mcs |= TIOCM_RTS;
  ioctl(fd, TIOCMODS, &mcs);
#endif
#ifdef _COHERENT
  ioctl(fd, TIOCSRTS, 0);
#endif

}

void LightweightSerial::serial_setparms(char *baudr, char par, char *bits,  int hwf, int swf)
{
  int spd = -1;
  unsigned int newbaud = 0;
  int bit = bits[0];
#ifdef _POSIX
  struct termios tty;
  tcgetattr(fd, &tty);
#else
  struct sgttyb tty;
  ioctl(fd, TIOCGETP, &tty);
#endif

  /* We generate mark and space parity ourself. */
  if(bit == '7' && (par == 'M' || par == 'S')) bit = '8';
  /* Check if 'baudr' is really a number */
  if((newbaud = (atol(baudr) / 100)) == 0 && baudr[0] != '0')
    newbaud = -1;

  switch(newbaud) {
  case 0:
#ifdef B0
    spd = B0;	   break;
#else
    spd = 0;  	   break;
#endif
  case 3:
    spd = B300;	   break;
  case 6:
    spd = B600;	   break;
  case 12:
    spd = B1200;   break;
  case 24:
    spd = B2400;   break;
  case 48:
    spd = B4800;   break;
  case 96:
    spd = B9600;   break;
#ifdef B19200
  case 192:
    spd = B19200;  break;
#else
#ifdef EXTA
  case 192:
    spd = EXTA;    break;
#else
  case 192:
    spd = B9600;   break;
#endif
#endif
#ifdef B38400
  case 384:
    spd = B38400;  break;
#else
#ifdef EXTB
  case 384:
    spd = EXTB;    break;
#else
  case 384:
    spd = B9600;   break;
#endif
#endif
#ifdef B57600
  case 576:
    spd = B57600;  break;
#endif
#ifdef B115200
  case 1152:
    spd = B115200; break;
#endif
#ifdef B500000
  case 5000:
    spd = B500000; break;
#endif
  }

#if defined (_BSD43) && !defined(_POSIX)
  if(spd != -1) tty.sg_ispeed = tty.sg_ospeed = spd;
  /* Number of bits is ignored */
  tty.sg_flags = RAW | TANDEM;
  if(par[0] == 'E')
    tty.sg_flags |= EVENP;
  else if(par[0] == 'O')
    tty.sg_flags |= ODDP;
  else
    tty.sg_flags |= PASS8 | ANYP;
  ioctl(fd, TIOCSETP, &tty);
#ifdef TIOCSDTR
  /* FIXME: huh? - MvS */
  ioctl(fd, TIOCSDTR, 0);
#endif
#endif

#if defined (_V7) && !defined(_POSIX)
  if(spd != -1) tty.sg_ispeed = tty.sg_ospeed = spd;
  tty.sg_flags = RAW;
  if(par == 'E')
    tty.sg_flags |= EVENP;
  else if(par == 'O')
    tty.sg_flags |= ODDP;
  ioctl(fd, TIOCSETP, &tty);
#endif

#ifdef _POSIX
  if(spd != -1) {
    cfsetospeed(&tty, (speed_t)spd);
    cfsetispeed(&tty, (speed_t)spd);
  }
  switch (bit) {
  case '5':
    tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS5;
    break;
  case '6':
    tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS6;
    break;
  case '7':
    tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS7;
    break;
  case '8':
  default:
    tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS8;
    break;
  }
  /* Set into raw, no echo mode */
#if !defined(_DGUX_SOURCE)
  tty.c_iflag &= ~(IGNBRK | IGNCR | INLCR | ICRNL | IUCLC |
                   IXANY | IXON | IXOFF | INPCK | ISTRIP);
  tty.c_iflag |= (BRKINT | IGNPAR);
  tty.c_oflag &= ~OPOST;
  tty.c_lflag = ~(ICANON | ISIG | ECHO | ECHONL | ECHOE | ECHOK | IEXTEN);
  tty.c_cflag |= CREAD | CRTSCTS;
#else /* Okay, this is better. XXX - Fix the above. */
  tty.c_iflag =  IGNBRK;
  tty.c_lflag = 0;
  tty.c_oflag = 0;
  tty.c_cflag |= CLOCAL | CREAD;
#endif
  tty.c_cc[VMIN] = 1;
  tty.c_cc[VTIME] = 5;

  /* Flow control. */
  if(hwf) {
    tty.c_cflag |= CRTSCTS;
    tty.c_cflag &= ~CLOCAL;
  }
  else {
    tty.c_cflag &= ~CRTSCTS;
    tty.c_cflag |= CLOCAL;
  }

  if(swf) {
    tty.c_iflag |= (IXON | IXOFF);
  }
  else {
    tty.c_iflag &= ~(IXON | IXOFF);
  }

  tty.c_cflag &= ~(PARENB | PARODD);

  if(par == 'E')
    tty.c_cflag |= PARENB;
  else if(par == 'O')
    tty.c_cflag |= PARODD;

  tcsetattr(fd, TCSANOW, &tty);

  serial_setrts();
#endif
}


long LightweightSerial::serial_numChars(void)
{
  long available = 0;

  if(ioctl(fd, FIONREAD, &available) == 0)
    return available;
  else  return -1;
}

int LightweightSerial::serial_ClearInputBuffer(void)
{
  unsigned char buffer[4096];
  int val = 0;

  val = serial_numChars();
  if(val > 0) read(fd, &buffer, val);
  return(val);
}

int LightweightSerial::serial_writen(char *buf, unsigned int nChars)
{
  int amountWritten = 0,count;

  count=(int)nChars;
  while(count > 0) {
    amountWritten = write(fd, buf, count);
    if(amountWritten < 0) {
      if(errno == EWOULDBLOCK) {
        fprintf(stderr, "\nWARNING: writeN: EWOULDBLOCK: trying again!\n");
      }
      else {
        return -1;
      }
    }
    else {
      count -= amountWritten;
      buf += amountWritten;
    }
  }
  return (int)nChars;
}


int LightweightSerial::serial_readn(char *buf, unsigned int nChars)
{
  int amountRead = 0, bytes_read = 0,count;
  struct timeval t;
  fd_set set;
  int err;

  if (happy==false) return -1;

  count=(int)nChars;
  while(count > 0) {
    t.tv_sec = 0;
    t.tv_usec = READ_TIMEOUT;
    FD_ZERO(&set);
    FD_SET(fd, &set);

    err = select(fd + 1, &set, NULL, NULL, &t);
    if(err == 0) return -2;

    amountRead = read(fd, buf, count);
    if(amountRead < 0 && errno != EWOULDBLOCK) return -1;
    else if(amountRead > 0) {
      bytes_read += amountRead;
      count -= amountRead;
      buf += amountRead;
    }
  }
  return bytes_read;
}


